/* @(#) tssa.c 1.16 96/08/26 12:02:37 */
/* /home/carrie/dist/tssa/SCCS/s.tssa.c */

/******************************************************************************/
/*  Copyright (C) 1996  Carrie Wager and Mark Segal.                          */
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 */
/*                                                                            */
/*  You may contact the authors via email: cwager@hsph.harvard.edu            */
/*  or mark@biostat.ucsf.edu, or by snail mail at:                            */
/*                                                                            */
/*  --------------------------------------------------------------------      */
/*  | Mark Segal                                                        |     */
/*  | Division of Biostatistics, University of California,  MU 420 West |     */
/*  | 500 Parnassus Avenue,  San Francisco,  CA  94143-0560             |     */
/*  --------------------------------------------------------------------      */
/******************************************************************************/


#include "tssa.h"
#include <S.h>
#ifdef WIN32
#  include <newredef.h>
#endif

/* global variable declarations */
static int      kN,		/* number of observations */
                kP,		/* number of covariates */
                kChoice,	/* choice for weights, should be enumerated type? */
                kMinBuc,	/* minimum bucket size */
                kMaxNodes;	/* maximum number of nodes, arbitrarily large
                				 * for array initialization only */

static double   kPropN,		/* minimum proportion of N allowed in bucket? */
                kEpsilon;	/* a small number */

static int      *gDelta, *gRiskSet, *gM, *gIndLefSam, *gN1, *gCaseNum;

static int      gCurNode, gNumNodes, gNumTerminal, gLength, gNumUnc;

static double   *gX, *gY, *gZ, *gW, *gWeights;

static struct node   **gTree;


int
tssa(tX, tY, tW, tDelta, tN, tP, tPropN, 
    tChoice, tEpsilon, tMaxNodes, tMinBuc, S_NodeNum, 
    S_SplitVar, S_SplitVal, S_N, S_SplitStat, 
    S_SplitMed, S_Devs, S_nnode, S_where)
  	double         *tX;
	double         *tY;
	double         *tW;
	int            *tDelta;
	int            *tN;
	int            *tP;
	double         *tPropN;

	int            *tChoice;
	double         *tEpsilon;
	int            *tMaxNodes;
	int            *tMinBuc;
	int            *S_NodeNum;

	int            *S_SplitVar;
	double         *S_SplitVal;
	int            *S_N;
	double         *S_SplitStat;

	double         *S_SplitMed;
	double         *S_Devs;
	int            *S_nnode;
	int            *S_where;
{
	int             i, j,s;

	gX = tX;
	gY = tY;
	gWeights = tW;
	gDelta = tDelta;

	kN = *tN;
	kP = *tP;
	kPropN = *tPropN;
	kChoice = *tChoice;
	kEpsilon = *tEpsilon;
	kMaxNodes = *tMaxNodes;
	kMinBuc = *tMinBuc;

	mySalloc(gRiskSet, kN, int);
	mySalloc(gM, kN, int);
	mySalloc(gIndLefSam, kN, int);
	mySalloc(gN1, kN, int);
	mySalloc(gCaseNum, kN, int);

	mySalloc(gW, kN, double);
	mySalloc(gZ, kN, double);

   /* variables needed for the reorder routine--save memory
      in Splus by allocating them outside the loop */

	mySalloc(gTree, kMaxNodes, struct node *);

	mySalloc(gTree[0], 1, struct node);

	gCurNode = gNumNodes = gNumTerminal = 0;

	/* initialization of case numbers */
	for (i = 0; i < kN; i++)
		gCaseNum[i] = i;

	/* initialization of root node details */
	gTree[0]->first = 0;
	gTree[0]->last = kN;
	gTree[0]->father = -1;

	/* main loop */
	while (gCurNode <= gNumNodes)
	{
		gLength = gTree[gCurNode]->last - gTree[gCurNode]->first + 1;
		if (gLength >= kMinBuc)
		{
			partit(&gNumNodes);
			gCurNode += 1;
		} else
		{
			terminal(gCurNode);
			gCurNode += 1;
			gNumTerminal += 1;
		}

	}

	/* recursive traversal of nodes to get order and numbering for S */
	gCurNode = 0;
	gTree[0]->snodenum = 1;
	traverse(0);

	/* now make return structures for S */
	for (i = 0; i <= gNumNodes; i++)
	{
		s = gTree[i]->sorder;
		S_NodeNum[s] = gTree[i]->snodenum;
		S_SplitVar[s] = gTree[i]->spl_coord;
		S_SplitVal[s] = gTree[i]->spl_value;
		S_N[s] = gTree[i]->last - gTree[i]->first;

		S_SplitStat[s] = gTree[i]->spl_stat;
		S_SplitMed[s] = gTree[i]->median;
		S_Devs[s] = gTree[i]->absdev;
	}
	*S_nnode = gNumNodes;
	for (j = 0; j < kN; j++)
	for (i = 0; i <= gNumNodes; i++)
      if((gTree[i]->first<=j 
          && (j<gTree[i]->last | (j<=gTree[i]->last && j==kN-1)) 
          && gTree[i]->left == -1) )
	      	S_where[gCaseNum[j]] = gTree[i]->snodenum;
   return 0;
}

void
partit(number)
int *number;
/* establishes whether a further split is worthwhile if so executes the
 * optimal split creating two children and defining their characteristics
 * predictors and response sorted according to current split coordinate */
{
	int             best_pred, split_pt, start, finish;
	double          best_stat, split_val;
   static int      firsttime=1;

	best_pred = 0;
	best_stat = 0;
	split_val = 0;
	split_pt = 0;

	start = gTree[gCurNode]->first;
	finish = gTree[gCurNode]->last;

/*	printf("About to call coord.\n"); */
	coord(start, finish, &best_pred, &best_stat, &split_val, &split_pt);
/*	printf("After calling coord.\n"); */

   if (*number >= (kMaxNodes - 2)) 
	{
      if(firsttime)
      {
         firsttime=0;
         printf("WARNING: maxnodes limit of %d was reached\n",kMaxNodes);
      }
		terminal(gCurNode);
		gNumTerminal += 1;
	}
	else if (best_stat > kEpsilon) 
	{
		gTree[gCurNode]->spl_coord = best_pred;
		gTree[gCurNode]->spl_value = split_val;
		gTree[gCurNode]->spl_stat = best_stat;
		gTree[gCurNode]->left = *number + 1;
		gTree[gCurNode]->right = *number + 2;

		*number += 1;
		mySalloc(gTree[*number], 1, struct node);
		gTree[*number]->father = gCurNode;
		gTree[*number]->first = gTree[gCurNode]->first;
		gTree[*number]->last = split_pt;

		*number += 1;
		mySalloc(gTree[*number], 1, struct node);
		gTree[*number]->father = gCurNode;
		gTree[*number]->first = split_pt;
		gTree[*number]->last = gTree[gCurNode]->last;

		reorder(best_pred, start, finish);

	} else
	{
		terminal(gCurNode);
		gNumTerminal += 1;
	}
}


void
coord(start, finish, pred, stat, val, pt)
	int             start;
	int             finish;
	int            *pred;
	double         *stat;
	double         *val;
	int            *pt;
{
	int             i, j, k; 
	double         twosam;

   int            *xjperm, *yperm;	
	double         *xjsort, *ysort;

/* I would like to declare these as local static pointers, and
   allocate only on the first function call, but that produces
   a bug where the memory for gTree[1]->last gets overwritten
   somehow when gY is copied into ysort--very strange!  
   So for now, the allocations here are somewhat excessive for 
   Splus, but not too bad.*/

	mySalloc(xjperm, kN, int);
	mySalloc(yperm, kN, int);
	mySalloc(xjsort, kN, double);
	mySalloc(ysort, kN, double);

/*	printf("Entering coord.\n"); */

	for (i = 0; i < kN; i++)
		gRiskSet[i] = gM[i] = (int) (gW[i] = (double) 0);

	twosam = 0;

	for (i = start; i < finish; i++)
		ysort[i] = gY[i];

	shell(ysort, start, finish, yperm);

	set(ysort, yperm, start, finish);
	ties(gCurNode);
	kaplan_meier(gCurNode);
	kmmedian(gCurNode);
	deviations(gCurNode);
	weights(kChoice, gCurNode);

	for (j = 0; j < kP; j++)
	{
		inital();

		for (i = start; i < finish; i++)
			xjsort[i] = GX(i, j);

		shell(xjsort, start, finish, xjperm);

		for (i = start; i < finish; i++)	/* update split point */
		{
			for (k = 0; k < gTree[gCurNode]->NumDistUnc; k++)
			{
				if (gZ[k] <= gY[xjperm[i]])
					gN1[k] += 1;

				if ((gDelta[xjperm[i]] == 1)
						&& (gZ[k] == gY[xjperm[i]]))
					gIndLefSam[k] += 1;
			}

			if ((i >= start + kMinBuc)
					&& (i <= (finish - kMinBuc)) && ((gNumUnc / (float) gLength) > kPropN)
					&& (xjsort[i] != xjsort[i + 1]))
				twosam = tarwar(i + 1, gCurNode);

			if (fabs(twosam) > *stat)
			{
				*stat = fabs(twosam);
				*pred = j;
				*pt = i + 1;
				*val = (xjsort[i] + xjsort[i + 1]) / 2.0;
			}
		}
	}
	Free(xjperm);
	Free(yperm);
	Free(xjsort);
	Free(ysort);
}

void
set(y, perm, first, last)
	double         *y;
	int            *perm;
	int             first;
	int             last;
/* initialize z and riskset vectors */
{
	int             i, j;

	j = 0;
	for (i = first; i < last; i++)
	{
		if (gDelta[perm[i]] == 1)
		{
			gZ[j] = y[i];
			gRiskSet[j] = last - i;
			j++;
		}
	}
	gNumUnc = j;
}

void
ties(nodenum)
	int             nodenum;
/* collect and count tied death times */
{
	int             count, tempm, i, j;

	count = 0;
	for (i = 0; ((i < gNumUnc) && (gZ[i] >= 0)); i++)
	{
		gM[i] = 1;

		for (j = i + 1; j < gNumUnc; j++)
		{
			if (gZ[i] == gZ[j])
				gM[i] += 1;
		}

		tempm = gM[i];
		while (tempm > 1)
		{
			count += 1;

			for (j = 1; i + j < gNumUnc - count; j++)
			{
				gZ[i + j] = gZ[i + j + 1];
				gRiskSet[i + j] = gRiskSet[i + j + 1];
			}

			for (j = gNumUnc - count - i; i + j < gNumUnc; j++)
				gZ[i + j] = -1;

			tempm -= 1;
		}
	}
	gTree[nodenum]->NumDistUnc = gNumUnc - count;
}

void
kaplan_meier(nodenum)
	int             nodenum;
/* computes kaplan-meier curve for terminal nodes */
{
	int             i;

	mySalloc(gTree[nodenum]->survcurv.S_axis, (gTree[nodenum]->NumDistUnc + 1), double);
	mySalloc(gTree[nodenum]->survcurv.t_axis, (gTree[nodenum]->NumDistUnc + 1), double);

	gTree[nodenum]->survcurv.t_axis[0] = 0;
	gTree[nodenum]->survcurv.S_axis[0] = 1;

	for (i = 1; i <= gTree[nodenum]->NumDistUnc; i++)
	{
		gTree[nodenum]->survcurv.t_axis[i] = gZ[i - 1];
		gTree[nodenum]->survcurv.S_axis[i] = gTree[nodenum]->survcurv.S_axis[i - 1]
			* (1 - (gM[i - 1] / (float) gRiskSet[i - 1]));
	}
}

void
kmmedian(nodenum)
	int             nodenum;

/* computes the (unsmoothed) kaplan-meier median */
{
	int             i;

	for (i = 0; i < gTree[nodenum]->NumDistUnc && gTree[nodenum]->survcurv.S_axis[i] > 0.5001; i++);

	if (gTree[nodenum]->survcurv.S_axis[i] >= 0.4999)
		gTree[nodenum]->median = gTree[nodenum]->survcurv.t_axis[i];
	else
		gTree[nodenum]->median = (gTree[nodenum]->survcurv.t_axis[i - 1] + gTree[nodenum]->survcurv.t_axis[i]) / 2.0;
}

void
deviations(nodenum)
	int             nodenum;

/* computes absolute deviations of node responses about k-m median */
{
	int             i;

	gTree[nodenum]->absdev = 0;

	for (i = gTree[nodenum]->first; i < gTree[nodenum]->last; i++)
		gTree[nodenum]->absdev += fabs(gY[i] - gTree[nodenum]->median);
}

void
weights(select, nodenum)
	int             select;
	int             nodenum;

/* specify weights via select for Tarone Ware class */
{
	int             j;

	switch (select)
	{
	case 1:			    /* w = 1 -> Mantel Haenzel */
		for (j = 0; j < gTree[nodenum]->NumDistUnc; j++)
			gW[j] = 1;
		break;

	case 2:			    /* w = riskset -> Gehan */
		for (j = 0; j < gTree[nodenum]->NumDistUnc; j++)
			gW[j] = gRiskSet[j];
		break;

	case 3:			    /* w = root(riskset) -> Tarone
					     * Ware */
		for (j = 0; j < gTree[nodenum]->NumDistUnc; j++)
			gW[j] = sqrt((double) gRiskSet[j]);
		break;

	case 4:			    /* w = failure time percentile ->
					     * Prentice */
		gW[0] = 1;
		for (j = 1; j < gTree[nodenum]->NumDistUnc; j++)
			gW[j] = gW[j - 1] * (gRiskSet[j - 1] / ((float) gRiskSet[j - 1] + 1));
		break;

	case 5:			    /* w = user specified */
		for (j = 0; j < gTree[nodenum]->NumDistUnc; j++)
                        gW[j] = gWeights[gCaseNum[j]];
		break;

	default:
		break;
	}
}

void
inital()
/* initialize left sample indicator and n1 vectors */
{
	int             i;

	for (i = 0; i < kN; i++)
	{
		gN1[i] = 0;
		gIndLefSam[i] = 0;
	}
}

void
shell(v, first, last, perm)
	double         *v;
	int             first;
	int             last;
	int            *perm;

/* sort v[first],...,v[last-1] into ascending order and give permutation
 * vector */
{
	int             gap, i, j, k, tperm;
	double          temp;

	for (k = first; k < last; k++)
		perm[k] = k;

	for (gap = (last - first) / 2; gap > 0; gap /= 2)
		for (i = first + gap; i < last; i++)
			for (j = i - gap; j >= first && v[j] > v[j + gap]; j -= gap)
			{
				temp = v[j];
				v[j] = v[j + gap];
				v[j + gap] = temp;
				tperm = perm[j];
				perm[j] = perm[j + gap];
				perm[j + gap] = tperm;
			}
}

double
tarwar(lef_sam_size, nodenum)
	int             lef_sam_size;
	int             nodenum;

/* compute Tarone-Ware statistic with weights */
{
	int             j;
	double          watediff, watevar, ratio;

	ratio = watediff = watevar = 0;
	j = 0;

	while ((gRiskSet[j] > 1) && (j < gTree[nodenum]->NumDistUnc))
	{
		ratio = gN1[j] / (float) gRiskSet[j];
		watediff += gW[j] * (gIndLefSam[j] - (ratio * gM[j]));
		watevar += gW[j] * gW[j] * (ratio * (1 - ratio) *
					    gM[j] * (gRiskSet[j] - gM[j])
					    / (gRiskSet[j] - 1));
		j++;
	}

	return (watediff / sqrt(watevar));
}

void
reorder(bestp, start, finish)
	int             bestp;
	int             start;
	int             finish;

/* reorders data (gY, gX, gDelta and case numbers) in accordance with best
 * predictor and over range of current node */
{
	int             i, j; 

   static int            *bestperm, *scrat_delta, *scrat_case;
	static double         *xbest, *scratx, *scraty;
   static int firsttime=1;

   if(firsttime){
      firsttime=0;
	   mySalloc(bestperm, kN, int);
	   mySalloc(scrat_delta, kN, int);
	   mySalloc(scrat_case, kN, int);

   	mySalloc(xbest, kN, double);
   	mySalloc(scratx, kN * kP, double);
   	mySalloc(scraty, kN, double);
   }

	for (i = start; i < finish; i++)
		xbest[i] = GX(i, bestp);

	shell(xbest, start, finish, bestperm);

	for (i = start; i < finish; i++)
	{
		scraty[i] = gY[i];
		scrat_delta[i] = gDelta[i];
		scrat_case[i] = gCaseNum[i];
		for (j = 0; j < kP; j++)
			scratx[(kP * i) + j] = GX(i, j);
	}

	for (i = start; i < finish; i++)
	{
		gY[i] = scraty[bestperm[i]];
		gDelta[i] = scrat_delta[bestperm[i]];
		gCaseNum[i] = scrat_case[bestperm[i]];

		for (j = 0; j < kP; j++)
			GX(i, j) = scratx[kP * bestperm[i] + j];
	}
}

void
terminal(current)
	int             current;
{
	gTree[current]->left = -1;
	gTree[current]->right = -1;
	gTree[current]->spl_coord = -1;
	gTree[current]->spl_value = -1;
	gTree[current]->spl_stat = -1;
}

void
traverse(nodenum)
	int             nodenum;
{
	if (nodenum > 0)
	{
		gTree[nodenum]->snodenum = (gTree[gTree[nodenum]->father]->snodenum) * 2;
		if (nodenum == gTree[gTree[nodenum]->father]->right)
			gTree[nodenum]->snodenum = gTree[nodenum]->snodenum + 1;
	}
	gTree[nodenum]->sorder = gCurNode;

	if (gTree[nodenum]->left != -1)
	{
		gCurNode++;
		traverse(gTree[nodenum]->left);
		gCurNode++;
		traverse(gTree[nodenum]->right);
	}
}

void
printkm(nodenum)
	int             nodenum;
/* writes kaplan-meier curve coordinates to files for all nodes */
{
	int             i, j, n;
	double         *step_t, *step_S;

	char            name[50];
	FILE           *fpkm;

	for (n = 0; n <= nodenum; n++)
	{

		sprintf(name, "node%d", n);
		fpkm = fopen(name, "w");

		mySalloc(step_t, (gTree[n]->NumDistUnc + 1) * 2, double);
		mySalloc(step_S, (gTree[n]->NumDistUnc + 1) * 2, double);

		step_t[0] = 0;

		for (i = 0; i < gTree[n]->NumDistUnc; i++)
			for (j = 0; j < 2; j++)
			{
				step_S[(2 * i) + j] = gTree[n]->survcurv.S_axis[i];
				step_t[(2 * i) + j + 1] = gTree[n]->survcurv.t_axis[i + 1];
			}

		step_S[(2 * gTree[n]->NumDistUnc)] = gTree[n]->survcurv.S_axis[gTree[n]->NumDistUnc];

		for (i = 0; i <= (2 * gTree[n]->NumDistUnc); i++)
			fprintf(fpkm, " %f  %f \n", step_t[i], step_S[i]);

		fclose(fpkm);
	}
}

void
detail(XNames)
   char **XNames;
/* writes out node details and reordered data */
{
	int             i, j;
	FILE           *fpdet;

	fpdet = fopen("details", "w");
	fprintf(fpdet, " %-30s \n", " Binary   Tree");
	fprintf(fpdet,
	  " %-8s %-8s %-8s %-8s %-10s %-10s %-10s %-10s %-10s %-8s %-8s \n",
		"node", "father", "left", "right", "spl-coord", "spl-value",
		"spl-stat", "median", "deviations", "first", "last");

	for (i = 0; i <= gNumNodes; i++)
		fprintf(fpdet,
			" %-8d %-8d %-8d %-8d %-10d %-10.3f %-10.3f %-10.3f %-10.3f %-8d %-8d \n",
			i, gTree[i]->father, gTree[i]->left,
			gTree[i]->right, gTree[i]->spl_coord, gTree[i]->spl_value, gTree[i]->spl_stat,
			gTree[i]->median, gTree[i]->absdev, gTree[i]->first, gTree[i]->last - 1);

	fprintf(fpdet, " %-30s \n", "Reordered   Data");
	fprintf(fpdet, " %-4s %-8s", "i", "case");

	for (j = 0; j < kP; j++)
		fprintf(fpdet, " %-8s", XNames[j]);

	fprintf(fpdet, " %-8s %-8s \n", "y", "delta");

	for (i = 0; i < kN; i++)
	{
		fprintf(fpdet, " %-4d %-8d", i, gCaseNum[i]);
		for (j = 0; j < kP; j++)
			fprintf(fpdet, " %-8.3f", GX(i, j));
		fprintf(fpdet, " %-8.3f %-8d \n", gY[i], gDelta[i]);
	}
	fclose(fpdet);
}

void
maxsubsplt()
/* This function performs bottom-up hierarchical pruning based on the maximum
 * split statistic for each subtree */
{
	int             i, parent, target, k, *node_term, *workperm;
	double         *mss, *work, tempmax;
	FILE           *fpnewseq;

	mySalloc(node_term, kMaxNodes, int);
	mySalloc(workperm, kMaxNodes, int);
	mySalloc(mss, kMaxNodes, double);
	mySalloc(work, kMaxNodes, double);

	fpnewseq = fopen("newseq", "w");

	for (i = 0; i <= gNumNodes; i++)
	{
		node_term[i] = 0;
		mss[i] = 0;
	}

	for (i = gNumNodes; i >= 0; i--)
	{
		if (gTree[i]->left == -1)
		{
			node_term[i] = 1;
			mss[i] = -1.0;
		} else
		{
			node_term[i] = node_term[gTree[i]->left] + node_term[gTree[i]->right];
			tempmax = (mss[gTree[i]->left] > mss[gTree[i]->right]) ?
				mss[gTree[i]->left] : mss[gTree[i]->right];
			mss[i] = (gTree[i]->spl_stat > tempmax) ? gTree[i]->spl_stat : tempmax;
		}
		work[i] = mss[i];
	}

	k = 1;

	fprintf(fpnewseq, "Tree 1  Term Nodes  %d  Subroot -1 Stat 0 \n", node_term[0]);

	shell(work, 0, gNumNodes, workperm);

	for (i = 0; i < gNumNodes; i++)
	{
		if (work[i] > -1)
		{
			target = workperm[i];
			parent = gTree[target]->father;

			while (parent >= 0 && mss[parent] == mss[target])
			{
				target = parent;
				parent = gTree[target]->father;
				i += 1;
			}

			k += 1;

			while (parent >= 0)
			{
				node_term[parent] -= (node_term[target] - 1);
				parent = gTree[parent]->father;
			}

			fprintf(fpnewseq, "Tree %d  Term Nodes  %d  Subroot %d Stat %f \n",
				k, node_term[0], target, mss[target]);
		}
		while (work[i] == work[i + 1])
			i += 1;
	}
	Free(node_term);
	Free(workperm);
	Free(mss);
	Free(work);
}
